/***************************************************************************
  Copyright 1996, 2002 Ronald S. Burkey.

  This file is part of Ron's Indexing Program (RIP).

  RIP is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  RIP is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with GutenMark; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  Filename:	huffman1.c
  Target:	MS-DOS
  Compiler:	Borland Turbo C 2.0
  PURPOSE:      CREATES HUFFMAN CODES FOR A FILE OF CHAR.
  MODS:		09/29/96 RSB	Adaped from IHUFFMAN.C
		10/07/96 RSB	Modified to generate context-based tables.
		01/08/02 RSB	Reformatted in GNU style and otherwise 
				cleaned up for first GPL'd release.
				
  This is a stand-alone utility program whose code does not get compiled into 
  rip.exe.  RIP compression operates with a static table of 
  language-dependent Huffman codes (though only English-language tables 
  were ever generated), and this program was used to initially generate 
  those tables.  The tables never changed later, because if they did change,
  no previously-compressed text could be accessed.  But you'd need this 
  program, for example, to create Huffman codes for a language previously
  unused by RIP.
  
  Huffman codes, in general, operate by encoding characters with variable-
  length sequences of bits.  More-frequently used characters have shorter
  bit-sequences and less-frequently used one have longer sequences.  
  The net result is that less bits are used, on average, than in the usual
  encodings.
  
  RIP improves on this somewhat by noting that the frequencies of appearance
  of characters are context-sensitive -- i.e., they depend on the characters
  that precede them.  Thus, 'u' is very likely to follow 'q', but is very
  UNlikely to follow 'u'.  So what we do is to choose a different table of
  Huffman codes for each preceding character.  	
  
  Unfortunately -- and I'm writing this 6 years after the Huffman tables 
  for RIP were actually generated -- there are various things about the
  generation process that were not recorded, and hence it cannot be 
  reproduced exactly.  First, huffman1.c is only one of a series of
  programs called huffmanX.c, and I'm not sure it's the one that was used.
  Secondly, I don't know what text file (or files) it was run on to 
  generate the tables.  Finally, the individual tables that were produced
  were not saved.  (The operation of rip.exe is such that when it 
  encounters the ASCII files produced by huffman1.c, it resaves them in 
  a single large file of binary format, for faster more efficient use, 
  and then uses the binary file thereafter.)  Oh, well; tough luck!
  As long as the binary file of Huffman codes exists, we really don't need
  to know how to reproduce it exactly, I guess.			
***************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>

char buffer[16384];

struct RECORD
{
  long count;
  struct RECORD *left;
  struct RECORD *right;
  int id;
};
struct RECORD counts[256];

double sum = 0.0;
long samples = 0;
int level = 0;
char s[50] = "";
void
traverse (struct RECORD *r)
{
  if (r->id > -1)
    {
      samples += r->count;
      sum += level * r->count;
      printf ("%4d:  %s\n", r->id, s);
      return;
    }
  if (r->left != NULL)
    {
      s[level++] = '0';
      s[level] = 0;
      traverse (r->left);
      level--;
    }
  if (r->right != NULL)
    {
      s[level++] = '1';
      s[level] = 0;
      traverse (r->right);
      level--;
    }
}

/* The significant thing about these characters is that they are
   overwhelmingly likely to be followed by a space. */
int
ispunctuation (char c)
{
  if (c == '.')
    return (1);
  if (c == '!')
    return (1);
  if (c == '?')
    return (1);
  if (c == ':')
    return (1);
  if (c == ';')
    return (1);
  if (c == ',')
    return (1);
  if (c == ')')
    return (1);
  return (0);
}

void
main (int argc, char *argv[])
{
  FILE *fin;
  unsigned char u;
  char inp;
  int i, j, n1, n2;
  long min1, min2;
  struct RECORD *r1, *r2;
  int numcounts, digit, alpha, punct, other, last;

  alpha = space = digit = punct = other = 0;

  for (i = 0; i < 256; i++)
    {
      counts[i].count = 0;
      counts[i].left = NULL;
      counts[i].right = NULL;
      counts[i].id = i;
    }

  /* get input file */
  if (argc < 2)
    {
      printf ("USAGE:\n\tHUFFMAN1 binary_integer_file [character]\n");
      printf ("The character parameter is A-Z,\n");
      printf ("punctuation (\".\"), decimal digit (\"0\"),\n");
      printf ("or other (\"*\").\n");
      exit (1);
    }
  if (argc > 2)
    {
      if (isalpha (*argv[2]))
	alpha = toupper (*argv[2]);
      else if (isdigit (*argv[2]))
	digit = 1;
      else if (ispunctuation (*argv[2]))
	punct = 1;
      else
	other = 1;
    }
  fin = fopen (argv[1], "rb");
  if (fin == NULL)
    {
      printf ("Input file not found.\n");
      exit (1);
    }
  setvbuf (fin, buffer, _IOFBF, sizeof (buffer));

  /* Make sure every character's got something in it */
  for (i = 0; i < 256; i++)
    counts[i].count = 1;

  /* take statistics */
  last = ' ';
  for (; 1 == fread (&inp, sizeof (inp), 1, fin); last = u)
    {
      u = inp;
      if (alpha)
	{
	  if (alpha != toupper (last))
	    continue;
	}
      else if (digit)
	{
	  if (!isdigit (last))
	    continue;
	}
      else if (punct)
	{
	  if (!ispunctuation (last))
	    continue;
	}
      else if (other)
	{
	  if (isalpha (last) || isdigit (last) || ispunctuation (last))
	    continue;
	}
      counts[u].count++;
    }

  numcounts = 256;

  /* informative printout */
  printf ("Criterion: ");
  if (alpha)
    printf ("%c\n", alpha);
  else if (digit)
    printf ("digit\n");
  else if (space)
    printf ("space\n");
  else if (punct)
    printf ("punctuation\n");
  else if (other)
    printf ("other\n");
  else
    printf ("everything\n");
  printf ("\nCounts:\n");
  for (i = 0; i < numcounts; i++)
    if (counts[i].count)
      printf ("%4d:  %ld\n", counts[i].id, counts[i].count);

  /* Huffman code */
  for (i = numcounts; i > 1; i--)
    {
      min1 = min2 = 10000000L;
      n1 = n2 = -1;
      for (j = 0; j < i; j++)
	if (counts[j].count < min2)
	  {
	    if (counts[j].count < min1)
	      {
		min2 = min1;
		n2 = n1;
		min1 = counts[j].count;
		n1 = j;
	      }
	    else
	      {
		min2 = counts[j].count;
		n2 = j;
	      }
	  }
      if (n2 == -1)
	{
	  printf ("Counts too large.\n");
	  exit (1);
	}
      r1 = (struct RECORD *) malloc (sizeof (struct RECORD));
      r2 = (struct RECORD *) malloc (sizeof (struct RECORD));
      if (r1 == NULL || r2 == NULL)
	{
	  printf ("Out of memory.\n");
	  exit (1);
	}
      *r1 = counts[n1];
      *r2 = counts[n2];
      counts[n1].count = r1->count + r2->count;
      counts[n1].right = r2;
      counts[n1].left = r1;
      counts[n1].id = -1;
      for (j = n2 + 1; j < i; j++)
	counts[j - 1] = counts[j];
    }

  printf ("\nCodes:\n");
  traverse (counts);

  printf ("\nAverage # bits per code:  %lg\n", sum / samples);
  fclose (fin);
  exit (0);
}


